package provide mock_ns 1.7.2
namespace eval mock_ns {
    namespace export mock_http
}

proc mock_ns::mock_http { args } {
    set global_vars [info vars ::*]
    variable mock_http_response_body ""
    
    qc::args $args -headers {} -- method url args
    set url_parts [qc::url_parts $url]
    
    if { [dict get $url_parts protocol] eq "" || [dict get $url_parts domain] eq "" } {
        error "Unsupported URL \"$url\". URLs mult be fully qualified with a protocol and domain"
    }

    try {
        # host header
        if { [dict get $url_parts port] eq "" } {
            multimap_set_first headers Host [dict get $url_parts domain]
        } else {
            multimap_set_first headers Host "[dict get $url_parts domain]:[dict get $url_parts port]"
        }
        
        ns_conn _set isconnected 1
        ns_conn _set method $method
        ns_conn _set protocol [dict get $url_parts protocol]
        ns_conn _set url [dict get $url_parts path]
        ns_conn _set headers [ns_set create headers {*}$headers]
        ns_conn _set version 1.1
        ns_conn _set peeraddr 127.0.0.1
        ns_conn _set flags 0
        ns_conn _set outputheaders [ns_set create outputheaders]
        ns_conn _set status "200"        

        # form variables, http_method
        if { $method in [list "GET" "HEAD"] } {
            # form data passed in URL query string
            set params [dict get $url_parts params]
            set form_data [qc::lapply url_decode $params]
            
            set http_method $method            
            
        } elseif { $method eq "POST" } {
            # form data passed in args
            set form_data $args

            set http_method $method            
            
        } else {
            # form data passed in args
            set form_data $args

            # method overloading
            multimap_set_first form_data _method $method
            set http_method POST           
            
        }
        ns_conn _set form [ns_set create form {*}$form_data]

        mock_http_marshal $http_method [dict get $url_parts path]

        return [dict create \
                    response_code [ns_conn status] \
                    response_headers [ns_set array [ns_conn outputheaders]] \
                    response_body $mock_http_response_body \
                   ]        
    } finally {        
        # unset any global variables created by mock_http
        unset -nocomplain {*}[lexclude [info vars ::*] {*}$global_vars]
        unset mock_http_response_body
        ns_conn _reset
    }
    
    return $body
}

proc mock_http_marshal { http_method url } {
    
    # preauth filters
    set result [_mock_http_marshal_ns_register_filter preauth $http_method $url]
    
    # postauth filters
    if { $result in [list "filter_ok" "filter_break"] } {
        set result [_mock_http_marshal_ns_register_filter postauth $http_method $url]
    }

    # procs
    if { $result ne "filter_return" } {
        _mock_http_marshal_ns_register_proc $http_method $url
    } 

    # trace filters
    _mock_http_marshal_ns_register_filter trace $http_method $url
}

proc _mock_http_marshal_ns_register_filter { when http_method url } {
    
    if { [nsv_exists "ns_register_filters" $when $http_method] } {
        foreach filter [nsv_get "ns_register_filters" $when $http_method] {
            if { [string match [dict get $filter url] $url] } {
                set result [[dict get $filter script] . {*}[dict get $filter args]]
                
                if { $result eq "filter_ok" } {
                    # continue with next filter
                    continue
                    
                } elseif { $result eq "filter_break" } {
                    # don't process any more filters of type $when
                    return "filter_break"
                    
                } elseif { $result eq "filter_return" } {
                    # don't run any more preauth, postauth or filters of type $when
                    # don't run procs registered for this method/URL                    
                    return "filter_return"
                    
                } else {
                    error "Unexpected filter return value \"$result\""
                    
                }
            }
        }
    }

    return "filter_ok"
}

proc _mock_http_marshal_ns_register_proc { http_method url } {
    set url [string trimright $url "/"]
    set matches {}
    if { [nsv_exists "ns_register_procs" $http_method] } {
        set data [nsv_get "ns_register_procs" $http_method]
        
        # check for -noinherit exact matches
        set noinherit true
        if { [dict exists $data $noinherit] } {
            foreach url_pattern [dict keys [dict get $data $noinherit]] {
                set pattern1 "$url_pattern"
                set pattern2 "${url_pattern}/"
                
                if { $url in [list $pattern1 pattern2] } {
                    lappend matches [dict_from \
                                         noinherit \
                                         url_pattern \
                                        ]
                }
            }
        }
        
        if { [llength $matches] == 0 } {
            # check for best inherit match
            set noinherit false
            if { [dict exists $data $noinherit] } {
                foreach url_pattern [dict keys [dict get $data $noinherit]] {
                    set segments [split [string trim $url_pattern "/"] "/"]
                    if { [llength $segments] > 1 } {
                        # escape all "*" characters that are not in the final URL segment
                        # to prevent glob matching
                        
                        set temp [string reverse $url_pattern]
                        set pattern1 [string reverse \
                                          [regsub \
                                               -all \
                                               -start [string first "/" $temp] \
                                               {\*} \
                                               $temp \
                                               {*\\} \
                                              ] \
                                         ]
                    } else {
                        set pattern1 $url_pattern
                    }

                    if {
                        [regexp {(^|/)\*$} $pattern1]
                        && [string match $pattern1 $url]
                    } {
                        # only "*" glob match char in last segment - allow inheritance match
                        lappend matches [dict_from \
                                             noinherit \
                                             url_pattern \
                                            ]                        
                    } elseif {
                              ![regexp {(^|/)\*$} $pattern1]
                              && ![string match "${pattern1}/*" $url] 
                              && [string match $pattern1 $url]
                          } {
                        # characters other than "*" in last segment - don't allow inheritance match
                        lappend matches [dict_from \
                                             noinherit \
                                             url_pattern \
                                            ]  
                    } elseif {
                              ![regexp {(^|/)[^/]*\*[^/]*$} $pattern1]
                              && [string match "${pattern1}/*" $url]
                          } {
                        # no "*" glob match char in last segment - allow inheritance match
                        lappend matches [dict_from \
                                             noinherit \
                                             url_pattern \
                                            ]                        
                    }
                }
            }
        }
    }
    
    if { [llength $matches] == 0 } {    
        error "No ns_register_proc match found for method \"$http_method\" and url \"$url\""
    } else {
        # select match with highest number of URL segments (prefer ones with no "*" glob char in last segment)
        set best_match [lindex $matches 0]
        set best_match_segments_count [llength [split \
                                                    [dict get $best_match url_pattern] \
                                                    "/"]]
        
        foreach match [lrange $matches 1 end] {
            dict2vars $match \
                url_pattern

            set segments_count [llength [split $url_pattern "/"]]

            if {
                $segments_count > $best_match_segments_count
                || (
                    $segments_count == $best_match_segments_count
                    && ![regexp {(^|/)[^/]*\*[^/]*$} $url_pattern]
                    )
            } {
                set best_match $match
                set best_match_segments_count $segments_count
            }
        }
        
        dict2vars $best_match \
            noinherit \
            url_pattern

        dict2vars [dict get $data $noinherit $url_pattern] \
            myproc \
            args
        
        return [$myproc {*}$args]
    }
}
