package provide mock_ns 1.7.2
namespace eval mock_ns {
    namespace export ns_conn
}
    
namespace eval mock_ns::ns_conn {
    namespace export \
        _set _reset auth authuser authpassword close flags form headers isconnected method outputheaders peeraddr \
        protocol location request status url version
    namespace ensemble create

    proc _coalesce {var_name default_value} {
        variable $var_name
        return [qc::coalesce $var_name $default_value]
    }
    
    proc _get {var_name} {
        variable $var_name
        return [set $var_name]
    }

    proc _reset {} {
        foreach var_name [info vars [namespace current]::*] {
            unset -nocomplain $var_name
        }
    }
    
    proc _set {var_name value} {
        variable $var_name
        return [set $var_name $value]
    }

    proc auth {} {
        set auth [ns_set create auth]
        
        if { [ns_set ifind [ns_conn headers] Authorization] >= 0 } {
            set authorization [ns_set iget [ns_conn headers] Authorization]

            set auth_method ""
            regexp {^(Basic|Bearer|Digest) +(.*)$} $authorization . auth_method token
            
            switch $auth_method {
                "Basic" {
                    lassign [split [::base64::decode $token] ":" ] \
                        username password

                    ns_set update $auth AuthMethod $auth_method
                    ns_set update $auth Username $username
                    ns_set update $auth Password $password
                }
                "Bearer" {
                    ns_set update $auth AuthMethod $auth_method
                    ns_set update $auth Token $token
                }
                "Digest" {
                    log "Digest Authorization scheme is not currently unsupported by mock-ns"  
                }
            }            
        }

        return $auth
    }

    proc authuser {} {
        set auth [ns_conn auth]

        if { [ns_set find $auth Username] >= 0 } {
            return [ns_set get $auth Username]
        } else {
            return ""
        }
    }

    proc authpassword {} {
        set auth [ns_conn auth]

        if { [ns_set find $auth Password] >= 0 } {
            return [ns_set get $auth Password]
        } else {
            return ""
        }
    }

    proc close {} {
        # update conn flags to indicate headers have been returns to client and connection is closed
        if { !([ns_conn flags] & 0x1) } {
            _set flags [expr {
                              [flags]
                              + 0x1
                          }]
        }
        if { !([ns_conn flags] & 0x10) } {
            _set flags [expr {
                              [flags]
                              + 0x10
                          }]
        }

        return [ns_conn flags]
    }
    
    proc flags {} {
        return [_coalesce flags 0]
    }
    
    proc form {} {
        return [_get form]
    }
    
    proc isconnected {} {
        return [_coalesce isconnected 0]
    }

    proc headers {} {
        return [_get headers]
    }
    
    proc \method {} {
        return [_get method]
    }    
    
    proc outputheaders {} {
        return [_get outputheaders]
    }

    proc peeraddr { args } {
        qc::args $args -source ? --
        if { ![info exists source] || $source eq "configured" } {
            return [_get peeraddr]
        }
        if { $source ni [list forwarded direct] } {
            error "ns_conn peeraddr: source $source not supported"
        }
        return [dict get [_get peeraddr] $source]
    }

    proc protocol {} {
        return [_get protocol]
    }

    proc location {} {
        variable ::mock_ns::ns_locationproc

        if { [info exists ns_locationproc] } {
            return [eval $ns_locationproc]
        } else {
            return [_get location]
        }
    }

    proc request {} {
        return "[method] [url] HTTP/[version]"
    }

    proc status { args } {
        if { [llength $args] > 1 } {
            error "Too many args for `ns_conn status ?status?`"
        }
        
        if { [llength $args] == 1 } {
            # set value 
            return [_set status [lindex $args 0]]
        } else {
            # get value
            return [_get status]
        }
    }
    
    proc url {} {
        return [_get url]
    }

    proc version {} {
        return [_get version]
    }
}
