package provide mock_ns 1.7.2
namespace eval mock_ns {
    namespace export ns_job
}

namespace eval mock_ns::ns_job {
    namespace export create joblist queue wait
    namespace ensemble create

    variable job_queues {}
    variable job_id 0

    proc create { args } {
        qc::args $args \
            -desc "" \
            -- \
            name \
            {maxthreads 4}

        variable job_queues

        if { [dict exists $job_queues $name] } {
            error "queue already exists: $name"
        }

        dict set job_queues $name [dict create \
                                       desc $desc \
                                       maxthreads $maxthreads \
                                       jobs {}]

        return $name
    }

    proc joblist { queue } {
        variable job_queues

        if { ![dict exists $job_queues $queue] } {
            error "no such queue: $queue"
        }

        return [dict values [dict get $job_queues $queue jobs]]
    }

    proc queue { args } {
        qc::args $args \
            -detached \
            -head \
            -jobid ? \
            -- \
            queue \
            script

        variable job_queues

        if { ![dict exists $job_queues $queue] } {
            error "no such queue: $queue"
        }

        if { [info exists jobid] } {
            if { [dict exists $job_queues $queue jobs $jobid] } {
                error "Job $jobid already exists"
            }
            set job_id $jobid
        } else {
            variable job_id
            incr job_id
        }

        if { [info exists detached] } {
            set type "detached"
        } else {
            set type "nondetached"
        }

        set job [dict create \
                     id $job_id \
                     state "scheduled" \
                     script $script \
                     results $script \
                     code "" \
                     type $type \
                     req "none" \
                     thread "" \
                     starttime "" \
                     endtime ""]
        dict set job_queues $queue jobs $job_id $job

        return $job_id
    }

    proc wait { args } {
        qc::args $args \
            -timeout ? \
            -- \
            queue \
            job_id

        variable job_queues

        if { ![dict exists $job_queues $queue] } {
            error "no such queue: $queue"
        }

        if { ![dict exists $job_queues $queue jobs $job_id] } {
            error "no such job: $job_id"
        }

        set job [dict get $job_queues $queue jobs $job_id]

        if { [dict get $job type] eq "detached" } {
            error "can't wait on detached job: $job_id"
        }

        if { [dict get $job req] eq "wait" } {
            error "can't wait on waited job: $job_id"
        }

        if { [dict get $job state] eq "done" } {
            return [dict get $job results]
        } else {
            dict set job_queues $queue jobs $job_id state "running"
            dict set job_queues $queue jobs $job_id req "wait"
            dict set job_queues $queue jobs $job_id starttime [clock seconds]

            set return_code [catch [dict get $job script] results options]

            dict set job_queues $queue jobs $job_id endtime [clock seconds]
            dict set job_queues $queue jobs $job_id state "done"
            dict set job_queues $queue jobs $job_id req "none"
            dict set job_queues $queue jobs $job_id results $results
            dict set job_queues $queue jobs $job_id code $return_code

            return $results
        }
    }
}
