package provide mock_ns 1.7.2
namespace eval mock_ns {
    namespace export ns_set
}

namespace eval mock_ns::ns_set {
    namespace export _reset array create find get ifind iget keys key put size unique update values value
    namespace ensemble create

    proc _reset {} {
        foreach var_name [info vars [namespace current]::*] {
            unset -nocomplain $var_name
        }
    }
    
    proc array { set_id } {
        variable $set_id
        return [list {*}[set $set_id]]
    }
    
    proc create { name args } {
        if { ([llength $args] % 2 ) != 0 } {
            error "Invalid args (key value pairs): $args"
        }

        variable index
        set set_id [incr0 index  1]
        variable $set_id
        set $set_id $args

        return $set_id
    }
    
    proc find { set_id key } {
        return [lsearch [keys $set_id] $key]
    }

    proc get { set_id key } {
        set index [find $set_id $key]

        return [value $set_id $index]        
    }

    proc ifind { set_id key } {
        return [lsearch -nocase [keys $set_id] $key]
        
    }
    
    proc iget { set_id key } {
        set index [ifind $set_id $key]
        
        return [value $set_id $index]      
    }
    
    proc keys { set_id {pattern "*"} } {
        variable $set_id
        set keys {}
        foreach {key value} [set $set_id] {
            if { [string match $pattern $key] } {
                lappend keys $key
            }
        }
        return $keys
    }
    
    proc key { set_id index } {
        return [lindex [keys $set_id] $index]
    }

    proc put { set_id key value } {
        variable $set_id
        set multimap [set $set_id]

        set index [size $set_id]
        lappend multimap $key $value
        
        set $set_id $multimap
        
        return $index
    }
    
    proc size { set_id } {
        variable $set_id
        return [expr {[llength [set $set_id]] /2}]
    }
    
    proc unique { set_id key } {
        if { [llength [keys $set_id $key]] <= 1 } {
            return true
        } else {
            return false
        }
    }

    proc update { set_id key value } {
        variable $set_id
        set multimap [set $set_id]
        set index [find $set_id $key]
        
        if { $index >= 0 } {
            lset multimap [expr {($index*2)+1}] $value
        } else {
            lappend multimap $key $value
        }
        
        set $set_id $multimap
        
        return $index
    }
    
    proc values { set_id {pattern "*"} } {
        variable $set_id
        set values {}
        foreach {key value} [set $set_id] {
            if { [string match $pattern value] } {
                lappend values $value
            }
        }
        return $values
    }
    
    proc value { set_id index } {
        return [lindex [values $set_id] $index]
    }
}
